local _G = _G

local function makeGlobalTable(name, major, minor, subName)
	subName = (subName or "") .. "Version"
	
	local table = _G[name]
	local majorName = subName .. "Major"
	local minorName = subName .. "Minor"
	if not table then
		--print("creating", name, subName, major, minor)
		local addon = { }
		addon[majorName] = major
		addon[minorName] = minor
		_G[name] = addon
		return addon
	end
		
	if not table[majorName]
		or table[majorName] < major 
		or (table[majorName] == major and table[minorName] < minor) then
		
		--print("upgrading", name, subName, "from", table[majorName], table[minorName], "to", major, minor)
		table[majorName] = major
		table[minorName] = minor
		return table
	end
	
	--print("skipping", name, subName, "from", table[majorName], table[minorName], "to", major, minor)
end

local u = makeGlobalTable("BittensUtilities", 1, 1)
if u == nil then
	return
end

local SlashCmdList = SlashCmdList
local pairs = pairs
local select = select
local type = type

-------------------------------------------------------------- Table Management
function u.MakeGlobalTable(name, major, minor, subName)
	return makeGlobalTable(name, major, minor, subName)
end

function u.GetOrMakeGlobalTable(name, major, minor, subName)
	local table = makeGlobalTable(name, major, minor, subName)
	if table then
		return table
	else
		return _G[name]
	end
end

function u.GetFromTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		outerTable = outerTable[key]
		if outerTable == nil then
			return nil
		end
	end
	return outerTable
end

function u.GetOrMakeTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if outerTable[key] == nil then
			outerTable[key] = {}
		end
		outerTable = outerTable[key]
	end
	return outerTable
end

function u.MagicallyIncrement(table, variable)
	local val = (table[variable] or 0) + 1
	table[variable] = val
	return val
end

---------------------------------------------------------------- Slash Commands
local slashHandlers = u.GetOrMakeGlobalTable("BittensSlashHandlers", 1, 1)
local subCommandHandlers =
	u.GetOrMakeGlobalTable("BittensSlashSubCommandHandlers", 1, 1)

local function getOrMakeSlashHandler(command)
	local handler = slashHandlers[command]
	if handler == nil then
		handler = function(userInput)
			local subCommand, rest = userInput:match("^(%S*)%s*(.-)$")
			if subCommand == nil 
				or subCommandHandlers[command][subCommand] == nil then
				
				if subCommandHandlers[command]["<no sub-command>"] then
					subCommandHandlers[command]["<no sub-command>"](userInput)
				else
					print("Possible sub-commands:")
					for sub, _ in pairs(subCommandHandlers[command]) do
						print("   ", sub)
					end
				end
			else
				subCommandHandlers[command][subCommand](rest)
			end
		end
		slashHandlers[command] = handler
	end
	return handler
end

-- "command" can actually be a table of commands, if you wish
function u.RegisterSlashCommand(command, subCommand, func)
	local allCommands
	if type(command) == "table" then
		allCommands = command
	else
		allCommands = { command }
	end
	local fullName = "BITTENS_UTILS_" .. allCommands[1]
	
	for i, command in pairs(allCommands) do
		_G["SLASH_" .. fullName .. i] = "/" .. command
	end
	
	SlashCmdList[fullName] = getOrMakeSlashHandler(command)
	
	u.GetOrMakeTable(subCommandHandlers, command)
		[subCommand or "<no sub-command>"] = func
end
